#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  menus.py
#  
#  Copyright 2017 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

"""
Menus are special submenus that act like modal dialogs.

They include glue code that automatically switches back to the previous submenu
after they are left. Note that this will cuase the :py:meth:`SubMenu.on_enter()`
method to be called again.

Since these menus are internally implemented as submenus, they are specific to their
:py:class:`Menu`\ , which must be active to be able to use the dialog.

Customization
-------------

Menus are customizable via several different means.

If you just want to change the appearance or label of a part of the menu, you can
use keyword arguments while initializing the class.
For example, setting the ``label_main`` argument to the string ``Hello World!`` the
main label or title of the dialog will now display ``Hello World!`` instead of its default
value.
What exact arguments are supported differs from dialog to dialog.

Note that sometimes specific labels are supported, but not used by default.
Just setting these to anything may cause GUI components to be rendered that
should not be there.

It is also possible for most of these values to be set on-the-fly via properties
on the object they belong to.

For example, the :py:attr:`DialogSubMenu.label_main` property may be set to change
the main label even while the dialog is active.

Note that the values accessible via keyword arguments and properties may differ.
This depends on the dialog implementing them.

For clarity, these keyword arguments will from now on be called "labels". This also
includes labels that are not strictly text, like the maximum value of a progressbar.
"""

__all__ = [
    "DialogSubMenu",
    "ConfirmSubMenu","TextSubMenu",
    "ProgressSubMenu","AdvancedProgressSubMenu",
    ]

import pyglet

from . import SubMenu
from . import text
from . import button
from . import slider

class DialogSubMenu(SubMenu):
    """
    Base Dialog Class.
    
    This class acts as a base class for all other dialog submenus.
    
    When the dialog is entered, the :py:attr:`prev_submenu` attribute will be set
    to the name of the previous submenu. This attribute is later used when exiting
    the dialog.
    
    Dialog submenus also support the basic actions used by all submenus, e.g.
    ``enter`` and ``exit``\ . Additionally, many dialogs also add actions for whenever
    a label is changed or the dialog is exited through a special means, e.g. pressing
    a specific button of multiple presented.
    
    If used by itself, it will display a text centered on the screen with a button
    below it. Clicking the button will cause the dialog to exit and also the
    additional ``click_ok`` action to be fired.
    
    The labels supported by default are ``label_main``\ , which defaults to ``Default Text``
    and is recommended to always be customized, and ``label_ok``\, which defaults to ``OK``
    and may be left as-is.
    
    Subclasses may override these defaults by setting the keys of the same name in the
    ``DEFAULT_LABELS`` class attribute. Note that any unchanged labels must also be declared
    when overwriting any labels, or they may not be displayed.
    
    Widgets and their initializers are stored in the :py:data:`WIDGETS` class attribute,
    see :py:meth:`add_widgets()` for more information.
    """
    DEFAULT_LABELS = {
        "label_main":"Default Text",
        "label_ok":"OK",
        }
    WIDGETS = {
        "label_main":"add_label_main",
        "label_ok":"add_btn_ok",
    }
    def __init__(self,name,menu,window,peng,
                borderstyle="oldshadow",
                **kwargs # for label_main etc.
                ):
        super(DialogSubMenu,self).__init__(name,menu,window,peng)
        
        self.prev_submenu = None
        self.borderstyle = borderstyle
        
        labels = {}
        labels.update(self.DEFAULT_LABELS)
        labels.update(kwargs)
        self.labels = kwargs
        self.kwargs = kwargs
        
        self.add_widgets(**labels)
    
    def add_widgets(self,**kwargs):
        """
        Called by the initializer to add all widgets.
        
        Widgets are discovered by searching through the :py:attr:`WIDGETS` class attribute.
        If a key in :py:attr:`WIDGETS` is also found in the keyword arguments and
        not none, the function with the name given in the value of the key will
        be called with its only argument being the value of the keyword argument.
        
        For more complex usage scenarios, it is also possible to override this method
        in a subclass, but the original method should always be called to ensure
        compatibility with classes relying on this feature.
        """
        for name,fname in self.WIDGETS.items():
            if name in kwargs and kwargs[name] is not None:
                assert hasattr(self,fname)
                assert callable(getattr(self,fname))
                getattr(self,fname)(kwargs[name])
    
    def add_label_main(self,label_main):
        """
        Adds the main label of the dialog.
        
        This widget can be triggered by setting the label ``label_main`` to a string.
        
        This widget will be centered on the screen.
        """
        # Main Label
        self.wlabel_main = text.Label("label_main",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,sh/2-bh/2),
                        size=[0,0],
                        label=label_main,
                        #multiline=True, # TODO: implement multine dialog
                        )
        self.wlabel_main.size = lambda sw,sh: (sw,self.wlabel_main._label.font_size)
        self.addWidget(self.wlabel_main)
    
    def add_btn_ok(self,label_ok):
        """
        Adds an OK button to allow the user to exit the dialog.
        
        This widget can be triggered by setting the label ``label_ok`` to a string.
        
        This widget will be mostly centered on the screen, but below the main label
        by the double of its height.
        """
        # OK Button
        self.wbtn_ok = button.Button("btn_ok",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,sh/2-bh/2-bh*2),
                        size=[0,0],
                        label=label_ok,
                        borderstyle=self.borderstyle
                        )
        self.wbtn_ok.size = lambda sw,sh: (self.wbtn_ok._label.font_size*8,self.wbtn_ok._label.font_size*2)
        self.addWidget(self.wbtn_ok)
        
        def f():
            self.doAction("click_ok")
            self.exitDialog()
        self.wbtn_ok.addAction("click",f)
        
    @property
    def label_main(self):
        """
        Property that proxies the ``label_main`` label.
        
        Setting this property will cause the ``label_main_change`` action to trigger.
        
        Note that trying to access this property if the widget is not used may cause
        an error.
        """
        # no check for initialized label, NameError should be good enough to debug
        return self.wlabel_main.label
    @label_main.setter
    def label_main(self,value):
        self.wlabel_main.label = value
        self.doAction("label_main_change")
    
    @property
    def label_ok(self):
        """
        Property that proxies the ``label_ok`` label.
        
        Setting this property will cause the ``label_ok_change`` action to trigger.
        
        Note that trying to access this property if the widget is not used may cause
        an error.
        """
        return self.wbtn_ok.label
    @label_ok.setter
    def label_ok(self,value):
        self.wbtn_ok.label = value
        self.doAction("label_ok_change")
    
    def on_enter(self,old):
        if self.menu.activeSubMenu==self.menu:
            raise RuntimeError("Cannot open a dialog twice")
        self.prev_submenu = old # name or None
    
    def exitDialog(self):
        """
        Helper method that exits the dialog.
        
        This method will cause the previously active submenu to activate.
        """
        if self.prev_submenu is not None:
            # change back to the previous submenu
            # could in theory form a stack if one dialog opens another
            self.menu.changeSubMenu(self.prev_submenu)
        self.prev_submenu = None
    
    def activate(self):
        """
        Helper method to enter the dialog.
        
        Calling this method will simply cause the dialog to become the active submenu.
        
        Note that is not necessary to call this method over :py:meth:`changeSubMenu()`\ , 
        as the storing of the previous submenu is done elsewhere.
        """
        # error checking done indirectly by on_enter
        # on_enter will be called automatically to store previous submenu
        self.menu.changeSubMenu(self.name)
    

class ConfirmSubMenu(DialogSubMenu):
    """
    Dialog that allows the user to confirm or cancel an action.
    
    By default, the OK button will be hidden and the ``label_main`` will be set
    to ``Are you sure?``\ .
    
    Clicking the confirm button will cause the ``confirm`` action to trigger, while
    the cancel button will cause the ``cancel`` action to trigger.
    """
    DEFAULT_LABELS = {
        "label_main":"Are you sure?",
        "label_confirm":"Confirm",
        "label_cancel":"Cancel",
        }
    WIDGETS = {
        **DialogSubMenu.WIDGETS,
        "label_confirm":"add_btn_confirm",
        "label_cancel":"add_btn_cancel",
        }
    
    def add_btn_confirm(self,label_confirm):
        """
        Adds a confirm button to let the user confirm whatever action they were presented with.
        
        This widget can be triggered by setting the label ``label_confirm`` to a string.
        
        This widget will be positioned slightly below the main label and to the left
        of the cancel button.
        """
        # Confirm Button
        self.wbtn_confirm = button.Button("btn_confirm",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw-4,sh/2-bh/2-bh*2),
                        size=[0,0],
                        label=label_confirm,
                        borderstyle=self.borderstyle
                        )
        self.wbtn_confirm.size = lambda sw,sh: (self.wbtn_confirm._label.font_size*8,self.wbtn_confirm._label.font_size*2)
        self.addWidget(self.wbtn_confirm)
        
        def f():
            self.doAction("confirm")
            self.exitDialog()
        self.wbtn_confirm.addAction("click",f)
    
    def add_btn_cancel(self,label_cancel):
        """
        Adds a cancel button to let the user cancel whatever choice they were given.
        
        This widget can be triggered by setting the label ``label_cancel`` to a string.
        
        This widget will be positioned slightly below the main label and to the right
        of the confirm button.
        """
        # Cancel Button
        self.wbtn_cancel = button.Button("btn_cancel",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2+4,sh/2-bh/2-bh*2),
                        size=[0,0],
                        label=label_cancel,
                        borderstyle=self.borderstyle
                        )
        self.wbtn_cancel.size = lambda sw,sh: (self.wbtn_cancel._label.font_size*8,self.wbtn_cancel._label.font_size*2)
        self.addWidget(self.wbtn_cancel)
        
        def f():
            self.doAction("cancel")
            self.exitDialog()
        self.wbtn_cancel.addAction("click",f)
    
    @property
    def label_confirm(self):
        """
        Property that proxies the ``label_confirm`` label.
        
        Setting this property will cause the ``label_confirm_change`` action to trigger.
        
        Note that trying to access this property if the widget is not used may cause
        an error.
        """
        return self.wbtn_confirm.label
    @label_confirm.setter
    def label_confirm(self,value):
        self.wbtn_confirm.label = value
        self.doAction("label_confirm_change")
    
    @property
    def label_cancel(self):
        """
        Property that proxies the ``label_cancel`` label.
        
        Setting this property will cause the ``label_cancel_change`` action to trigger.
        
        Note that trying to access this property if the widget is not used may cause
        an error.
        """
        return self.wbtn_cancel.label
    @label_cancel.setter
    def label_cancel(self,value):
        self.wbtn_cancel.label = value
        self.doAction("label_cancel_change")

class TextSubMenu(DialogSubMenu):
    """
    Dialog without user interaction that can automatically exit after a certain amount of time.
    
    This dialog accepts the ``timeout`` keyword argument, which may be set to any
    time in seconds to delay before exiting the dialog. A value of ``-1`` will cause
    the dialog to never exit on its own.
    
    Note that the user will not be able to exit this dialog and may believe the program
    is hanging if not assured otherwise. It is thus recommended to use the :py:class:`ProgressSubMenu`
    dialog instead, especially for long-running operations.
    """
    DEFAULT_LABELS = {
        "label_main":"Default Text",
        # no button needed, timer does the rest
        }
    def __init__(self,name,menu,window,peng,
                borderstyle="oldshadow",
                timeout=10,
                **kwargs
                ):
        super(TextSubMenu,self).__init__(name,menu,window,peng,borderstyle,**kwargs)
        self.timeout = timeout
    
    def on_enter(self,old):
        super(TextSubMenu,self).on_enter(old)
        
        if self.timeout!=-1:
            pyglet.clock.schedule_once(lambda dt:self.exitDialog(),self.timeout)

class ProgressSubMenu(DialogSubMenu):
    """
    Dialog without user interaction displaying a progressbar.
    
    By default, the progressbar will range from 0-100, effectively a percentage.
    
    The :py:attr:`auto_exit` attribute may be set to control whether or not the dialog
    will exit automatically when the maximum value is reached.
    """
    DEFAULT_LABELS = {
        "label_main":"Loading...",
        "label_progressbar":"{percent:.1}%",
        # TODO: actually implement the progress_* labels
        "progress_n":0, # should be updated on-the-fly through property progress_n
        "progress_nmin":0,
        "progress_nmax":100, # basically equal to percentages
        }
    WIDGETS = {
        **DialogSubMenu.WIDGETS,
        "label_progressbar":"add_progressbar",
        }
    auto_exit = False
    """
    Controls whether or not the dialog will exit automatically after the maximum
    value has been reached.
    """
    
    def add_progressbar(self,label_progressbar):
        """
        Adds a progressbar and label displaying the progress within a certain task.
        
        This widget can be triggered by setting the label ``label_progressbar`` to
        a string.
        
        The progressbar will be displayed centered and below the main label.
        The progress label will be displayed within the progressbar.
        
        The label of the progressbar may be a string containing formatting codes
        which will be resolved via the ``format()`` method.
        
        Currently, there are six keys available:
        
        ``n`` and ``value`` are the current progress rounded to 4 decimal places.
        
        ``nmin`` is the minimum progress value rounded to 4 decimal places.
        
        ``nmax`` is the maximum progress value rounded to 4 decimal places.
        
        ``p`` and ``percent`` are the percentage value that the progressbar is completed
        rounded to 4 decimal places.
        
        By default, the progressbar label will be ``{percent}%`` displaying the percentage
        the progressbar is complete.
        """
        # Progressbar
        self.wprogressbar = slider.Progressbar("progressbar",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,self.wlabel_main.pos[1]-bh*1.5),
                        size=[0,0],
                        #label=label_progressbar # TODO: add label
                        borderstyle=self.borderstyle
                        )
        self.addWidget(self.wprogressbar)
        
        # Progress Label
        self.wprogresslabel = text.Label("progresslabel",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,self.wprogressbar.pos[1]+8),
                        size=[0,0],
                        label="", # set by update_progressbar()
                        )
        self.wprogresslabel.size = lambda sw,sh: (sw,self.wprogresslabel._label.font_size)
        self.addWidget(self.wprogresslabel)
        
        self.wprogressbar.size = lambda sw,sh: (sw*0.8,self.wprogresslabel._label.font_size+10)
        
        self._label_progressbar = label_progressbar
        
        if getattr(label_progressbar,"_dynamic",False):
            def f():
                self.label_progressbar = str(label_progressbar)
            self.peng.i18n.addAction("setlang",f)
        
        self.wprogressbar.addAction("progresschange",self.update_progressbar)
        
        self.update_progressbar()
    
    def update_progressbar(self):
        """
        Updates the progressbar by re-calculating the label.
        
        It is not required to manually call this method since setting any of the
        properties of this class will automatically trigger a re-calculation.
        """
        n,nmin,nmax = self.wprogressbar.n,self.wprogressbar.nmin,self.wprogressbar.nmax
        if (nmax-nmin)==0:
            percent = 0 # prevents ZeroDivisionError
        else:
            percent = max(min((n-nmin)/(nmax-nmin),1.),0.)*100
        dat = {"value":round(n,4),"n":round(n,4),"nmin":round(nmin,4),"nmax":round(nmax,4),"percent":round(percent,4),"p":round(percent,4)}
        txt = self._label_progressbar.format(**dat)
        self.wprogresslabel.label = txt
    
    @property
    def progress_n(self):
        """
        Property that proxies the ``progress_n`` label.
        
        Setting this property will cause the progressbar label to be recalculated.
        
        Additionally, if the supplied value is higher than the maximum value and
        :py:attr:`auto_exit` is true, the dialog will exit.
        """
        return self.wprogressbar.n
    @progress_n.setter
    def progress_n(self,value):
        self.wprogressbar.n = value
        self.update_progressbar()
        if self.auto_exit:
            if self.wprogressbar.n>=self.wprogressbar.nmax:
                self.exitDialog()
    
    @property
    def progress_nmin(self):
        """
        Property that proxies the ``progress_nmin`` label.
        
        Setting this property will cause the progressbar label to be recalculated.
        
        Note that setting this property if the widget has not been initialized may
        cause various errors to occur.
        """
        return self.wprogressbar.nmin
    @progress_nmin.setter
    def progress_nmin(self,value):
        self.wprogressbar.nmin = value
        self.update_progressbar()
    
    @property
    def progress_nmax(self):
        """
        Property that proxies the ``progress_nmax`` label.
        
        Setting this property will cause the progressbar label to be recalculated.
        
        Note that setting this property if the widget has not been initialized may
        cause various errors to occur.
        """
        return self.wprogressbar.nmax
    @progress_nmax.setter
    def progress_nmax(self,value):
        self.wprogressbar.nmax = value
        self.update_progressbar()
    
    @property
    def label_progressbar(self):
        """
        Property that proxies the ``label_progressbar`` label.
        
        Setting this property will cause the progressbar label to be recalculated.
        
        Note that setting this property if the widget has not been initialized may
        cause various errors to occur.
        """
        return self.wprogresslabel.label
    @label_progressbar.setter
    def label_progressbar(self,value):
        self._label_progressbar = value
        self.update_progressbar()
    

class AdvancedProgressSubMenu(ProgressSubMenu):
    def add_progressbar(self,label_progressbar):
        """
        Adds a progressbar and label displaying the progress within a certain task.
        
        This widget can be triggered by setting the label ``label_progressbar`` to
        a string.
        
        The progressbar will be displayed centered and below the main label.
        The progress label will be displayed within the progressbar.
        
        The label of the progressbar may be a string containing formatting codes
        which will be resolved via the ``format()`` method.
        
        Currently, there are six keys available:
        
        ``n`` and ``value`` are the current progress rounded to 4 decimal places.
        
        ``nmin`` is the minimum progress value rounded to 4 decimal places.
        
        ``nmax`` is the maximum progress value rounded to 4 decimal places.
        
        ``p`` and ``percent`` are the percentage value that the progressbar is completed
        rounded to 4 decimal places.
        
        By default, the progressbar label will be ``{percent}%`` displaying the percentage
        the progressbar is complete.
        """
        # Progressbar
        self.wprogressbar = slider.AdvancedProgressbar("progressbar",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,self.wlabel_main.pos[1]-bh*1.5),
                        size=[0,0],
                        #label=label_progressbar # TODO: add label
                        borderstyle=self.borderstyle
                        )
        self.addWidget(self.wprogressbar)
        
        # Progress Label
        self.wprogresslabel = text.Label("progresslabel",self,self.window,self.peng,
                        pos=lambda sw,sh, bw,bh: (sw/2-bw/2,self.wprogressbar.pos[1]+8),
                        size=[0,0],
                        label="", # set by update_progressbar()
                        )
        self.wprogresslabel.size = lambda sw,sh: (sw,self.wprogresslabel._label.font_size)
        self.addWidget(self.wprogresslabel)
        
        self.wprogressbar.size = lambda sw,sh: (sw*0.8,self.wprogresslabel._label.font_size+10)
        
        self._label_progressbar = label_progressbar
        
        if getattr(label_progressbar,"_dynamic",False):
            def f():
                self.label_progressbar = str(label_progressbar)
            self.peng.i18n.addAction("setlang",f)
        
        self.wprogressbar.addAction("progresschange",self.update_progressbar)
        
        self.update_progressbar()
    
    def addCategory(self,*args,**kwargs):
        """
        Proxy for :py:meth:`~peng3d.gui.slider.AdvancedProgressbar.addCategory()`\ .
        """
        return self.wprogressbar.addCategory(*args,**kwargs)
    def updateCategory(self,*args,**kwargs):
        """
        Proxy for :py:meth:`~peng3d.gui.slider.AdvancedProgressbar.updateCategory()`\ .
        """
        return self.wprogressbar.updateCategory(*args,**kwargs)
    def deleteCategory(self,*args,**kwargs):
        """
        Proxy for :py:meth:`~peng3d.gui.slider.AdvancedProgressbar.deleteCategory()`\ .
        """
        return self.wprogressbar.deleteCategory(*args,**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

import warnings

# needed for py27 compat
import weakref as _weakref

class _WeakMethod(object):
    """A callable object. Takes one argument to init: 'object.method'.
    Once created, call this object -- MyWeakMethod() -- 
    and pass args/kwargs as you normally would.
    """
    def __init__(self, object_dot_method):
        self.target = _weakref.proxy(object_dot_method.__self__)
        self.method = _weakref.proxy(object_dot_method.__func__)
        ###Older versions of Python can use 'im_self' and 'im_func' in place of '__self__' and '__func__' respectively

    def __call__(self):
        def f(*args,**kwargs):
            return self.method(self.target, *args, **kwargs)
        return f
_weakref.WeakMethod = _WeakMethod

# needed for bidict<0.14.0 compat
import bidict

if not hasattr(bidict,"OrderedBidict"):
    bidict.OrderedBidict = bidict.orderedbidict
    warnings.warn("Support for bidict<0.14.0 will be dropped in a future version",DeprecationWarning)


# These imports are here for convenience

from .peng import *
#from .window import *
from .layer import *
from .menu import *
from .camera import *
from .config import *
from .version import *
from .world import *
from .actor.player import *
from .actor import *
from .gui import *
from .gui.widgets import *
from .gui.button import *
from .gui.slider import *
from .gui.text import *
from .gui.container import *
from .gui.menus import *
from .resource import *
from .i18n import *
from .model import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  window.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["PengWindow"]

import math
import traceback
import weakref
import inspect

import pyglet
from pyglet.gl import *
from pyglet.window import key

from . import config, camera

ICON_SIZES = [16,24,32,48,64,128,256,512]

class PengWindow(pyglet.window.Window):
    """
    Main window class for peng3d and subclass of :py:class:`pyglet.window.Window()`\ .
    
    This class should not be instantiated directly, use the :py:meth:`Peng.createWindow()` method.
    """
    
    def __init__(self,peng,*args,**kwargs):
        if peng.cfg["graphics.stencil.enable"]:
            glconfig = pyglet.gl.Config(stencil_size=peng.cfg["graphics.stencil.bits"])
            kwargs["config"]=glconfig
        super(PengWindow,self).__init__(*args,**kwargs)
        self.peng = peng
        
        self.exclusive = False
        self.started = False
        self.exclusive = False
        
        self.menus = {}
        self.activeMenu = None
        
        self.cfg = config.Config({},defaults=peng.cfg)
        self.eventHandlers = {}
        
        self._setup = False
        def on_key_press(symbol, modifiers):
            if symbol == key.ESCAPE:
                return pyglet.event.EVENT_HANDLED
        self.push_handlers(on_key_press) # to stop the auto-exit on escape
        
        self.peng.sendEvent("peng3d:window.create",{"peng":self.peng,"window":self})
    def setup(self):
        """
        Sets up the OpenGL state.
        
        This method should be called once after the config has been created and before the main loop is started.
        You should not need to manually call this method, as it is automatically called by :py:meth:`run()`\ .
        
        Repeatedly calling this method has no effects.
        """
        if self._setup:
            return
        self._setup = True
        
        # Ensures that default values are supplied
        #self.cleanConfig()
        
        # Sets min/max window size, if specified
        if self.cfg["graphics.min_size"] is not None:
            self.set_minimum_size(*self.cfg["graphics.min_size"])
        if self.cfg["graphics.max_size"] is not None:
            self.set_maximum_size(*self.cfg["graphics.max_size"])
        
        # Sets up basic OpenGL state
        glClearColor(*self.cfg["graphics.clearColor"])
        
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_BLEND)
        
        glShadeModel(GL_SMOOTH)
        
        if self.cfg["graphics.fogSettings"]["enable"]:
            self.setupFog()
        if self.cfg["graphics.lightSettings"]["enable"]:
            self.setupLight()
    def setupFog(self):
        """
        Sets the fog system up.
        
        The specific options available are documented under :confval:`graphics.fogSettings`\ .
        """
        fogcfg = self.cfg["graphics.fogSettings"]
        if not fogcfg["enable"]:
            return
        
        glEnable(GL_FOG)
        
        if fogcfg["color"] is None:
            fogcfg["color"] = self.cfg["graphics.clearColor"]
        # Set the fog color.
        glFogfv(GL_FOG_COLOR, (GLfloat * 4)(*fogcfg["color"]))
        # Set the performance hint.
        glHint(GL_FOG_HINT, GL_DONT_CARE) # TODO: add customization, including headless support
        # Specify the equation used to compute the blending factor.
        glFogi(GL_FOG_MODE, GL_LINEAR)
        # How close and far away fog starts and ends. The closer the start and end,
        # the denser the fog in the fog range.
        glFogf(GL_FOG_START, fogcfg["start"])
        glFogf(GL_FOG_END, fogcfg["end"])
        
    def setupLight(self):
        """
        Sets the light system up.
        
        The specific options available are documented under :confval:`graphics.lightSettings`\ .
        
        Note that this feature is currently not implemented.
        """
        raise NotImplementedError("Currently not implemented")
    
    def run(self,evloop=None):
        """
        Runs the application in the current thread.
        
        This method should not be called directly, especially when using multiple windows, use :py:meth:`Peng.run()` instead.
        
        Note that this method is blocking as rendering needs to happen in the main thread.
        It is thus recommendable to run your game logic in another thread that should be started before calling this method.
        
        ``evloop`` may optionally be a subclass of :py:class:`pyglet.app.base.EventLoop` to replace the default event loop.
        """
        self.setup()
        if evloop is not None:
            pyglet.app.event_loop = evloop
        pyglet.app.run() # This currently just calls the basic pyglet main loop, maybe implement custom main loop for more control
    """def cleanConfig(self):
        ########## Reset quotation marks if uncommenting
        Sets default values for various config values.
        
        .. todo::
           
           Use an defaultdict or similiar instead.
        ########## End quotation marks
        # Various default values are set in this method, this should really be replaced with something more easy to use and robust
        # A possible replacement could be a defaultdict or similiar
        # Update: now replaced by :py:class:`Config()`\ .
        
        # OpenGL configs
        self.cfg["graphics.clearColor"] = self.cfg.get("graphics.clearColor",(0.,0.,0.,1.))
        self.cfg["graphics.wireframe"] = self.cfg.get("graphics.wireframe",False)
        self.cfg["graphics.fieldofview"] = self.cfg.get("graphics.fieldofview",65.0)
        self.cfg["graphics.nearclip"] = self.cfg.get("graphics.nearclip",0.1)
        self.cfg["graphics.farclip"] = self.cfg.get("graphics.farclip",10000) # It's over 9000!
        
        # OpenGL - Fog
        self.cfg["graphics.fogSettings"] = self.cfg.get("graphics.fogSettings",CFG_FOG_DEFAULT)
        self.cfg["graphics.fogSettings"]["enable"] = self.cfg["graphics.fogSettings"].get("enable",False)
        if self.cfg["graphics.fogSettings"]["enable"]:
            pass
        
        # OpenGL - Light
        self.cfg["graphics.lightSettings"] = self.cfg.get("graphics.lightSettings",CFG_LIGHT_DEFAULT)
        self.cfg["graphics.lightSettings"]["enable"] = self.cfg["graphics.lightSettings"].get("enable",False)
        if self.cfg["graphics.lightSettings"]["enable"]:
            pass
        
        # Other configs
        
        return
    """
    # Various methods
    def changeMenu(self,menu):
        """
        Changes to the given menu.
        
        ``menu`` must be a valid menu name that is currently known.
        
        .. versionchanged:: 1.2a1
           
           The push/pop handlers have been deprecated in favor of the new :py:meth:`Menu.on_enter() <peng3d.menu.Menu.on_enter>`\ , :py:meth:`Menu.on_exit() <peng3d.menu.Menu.on_exit>`\ , etc. events.
        """
        if menu not in self.menus:
            raise ValueError("Menu %s does not exist!"%menu)
        elif menu == self.activeMenu:
            return # Ignore double menu activation to prevent bugs in menu initializer
        old = self.activeMenu
        self.activeMenu = menu
        if old is not None:
            self.menus[old].on_exit(menu)
            self.menus[old].doAction("exit")
            #self.pop_handlers()
        self.menu.on_enter(old)
        self.menu.doAction("enter")
        self.peng.sendEvent("peng3d:window.menu.change",{"peng":self.peng,"window":self,"old":old,"menu":menu})
        #self.push_handlers(self.menu)
    def addMenu(self,menu):
        """
        Adds a menu to the list of menus.
        """
        # If there is no menu selected currently, this menu will automatically be made active.
        # Add the line above to the docstring if fixed
        self.menus[menu.name]=menu
        self.peng.sendEvent("peng3d:window.menu.add",{"peng":self.peng,"window":self,"menu":menu})
        #if self.activeMenu is None:
        #    self.changeMenu(menu.name)
        # currently disabled because of a bug with adding layers
    
    def setIcons(self,icons):
        if isinstance(icons,list) or isinstance(icons,tuple):
            # should be a list of resource names
            rlist = icons
        else:
            # We were given a string, need to expand it first
            rlist = []
            for isize in ICON_SIZES:
                rname = icons.format(size=isize)
                if self.peng.rsrcMgr.resourceExists(rname,".png"):
                    # Resource exists, add it to list
                    rlist.append(rname)
        
        ilist = []
        for rname in rlist:
            ilist.append(pyglet.image.load(self.peng.rsrcMgr.resourceNameToPath(rname,".png")))
        
        if len(ilist)!=0:
            self.set_icon(*ilist)
    
    # Event handlers
    def on_draw(self):
        """
        Clears the screen and draws the currently active menu.
        """
        self.clear()
        self.menu.draw()
    
    def dispatch_event(self,event_type,*args):
        """
        Internal event handling method.
        
        This method extends the behavior inherited from :py:meth:`pyglet.window.Window.dispatch_event()` by calling the various :py:meth:`handleEvent()` methods.
        
        By default, :py:meth:`Peng.handleEvent()`\ , :py:meth:`handleEvent()` and :py:meth:`Menu.handleEvent()` are called in this order to handle events.
        
        Note that some events may not be handled by all handlers during early startup.
        """
        super(PengWindow,self).dispatch_event(event_type,*args)
        try:
            p = self.peng
            m = self.menu
        except AttributeError:
            # To prevent early startup errors
            if hasattr(self,"peng") and self.peng.cfg["debug.events.logerr"]:
                print("Error:")
                traceback.print_exc()
            return
        p.handleEvent(event_type,args,self)
        self.handleEvent(event_type,args)
        m.handleEvent(event_type,args)
    
    def handleEvent(self,event_type,args,window=None):
        args = list(args)
        #if window is not None:
        #    args.append(window)
        if event_type in self.eventHandlers:
            for whandler in self.eventHandlers[event_type]:
                # This allows for proper collection of deleted handler methods by using weak references
                handler = whandler()
                if handler is None:
                    del self.eventHandlers[event_type][self.eventHandlers[event_type].index(whandler)]
                handler(*args)
    handleEvent.__noautodoc__ = True
    
    def registerEventHandler(self,event_type,handler):
        if self.peng.cfg["debug.events.register"]:
            print("Registered Event: %s Handler: %s"%(event_type,handler))
        if event_type not in self.eventHandlers:
            self.eventHandlers[event_type]=[]
        # Only a weak reference is kept
        if inspect.ismethod(handler):
            handler = weakref.WeakMethod(handler)
        else:
            handler = weakref.ref(handler)
        self.eventHandlers[event_type].append(handler)
    
    # Properties/Proxies for various things
    
    # Proxy for self.menus[self.activeMenu]
    @property
    def menu(self):
        """
        Property for accessing the currently active menu.
        
        Always equals ``self.menus[self.activeMenu]``\ .
        
        This property is read-only.
        """
        return self.menus[self.activeMenu]
    
    # Utility methods
    
    def toggle_exclusivity(self,override=None):
        """
        Toggles mouse exclusivity via pyglet's :py:meth:`set_exclusive_mouse()` method.
        
        If ``override`` is given, it will be used instead.
        
        You may also read the current exclusivity state via :py:attr:`exclusive`\ .
        """
        if override is not None:
            new = override
        else:
            new = not self.exclusive
        self.exclusive = new
        self.set_exclusive_mouse(self.exclusive)
        self.peng.sendEvent("peng3d:window.toggle_exclusive",{"peng":self.peng,"window":self,"exclusive":self.exclusive})
    
    def set2d(self):
        """
        Configures OpenGL to draw in 2D.
        
        Note that wireframe mode is always disabled in 2D-Mode, but can be re-enabled by calling ``glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)``\ .
        """
        # Light
        
        glDisable(GL_LIGHTING)
        
        # To avoid accidental wireframe GUIs and fonts
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL)
        
        width, height = self.get_size()
        glDisable(GL_DEPTH_TEST)
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, width, 0, height, -1, 1)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
    def set3d(self,cam):
        """
        Configures OpenGL to draw in 3D.
        
        This method also applies the correct rotation and translation as set in the supplied camera ``cam``\ .
        It is discouraged to use :py:func:`glTranslatef()` or :py:func:`glRotatef()` directly as this may cause visual glitches.
        
        If you need to configure any of the standard parameters, see the docs about :doc:`/configoption`\ .
        
        The :confval:`graphics.wireframe` config value can be used to enable a wireframe mode, useful for debugging visual glitches.
        """
        if not isinstance(cam,camera.Camera):
            raise TypeError("cam is not of type Camera!")
        
        # Light
        
        #glEnable(GL_LIGHTING)
        
        if self.cfg["graphics.wireframe"]:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        
        width, height = self.get_size()
        glEnable(GL_DEPTH_TEST)
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(self.cfg["graphics.fieldofview"], width / float(height), self.cfg["graphics.nearclip"], self.cfg["graphics.farclip"]) # default 60
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        x, y = cam.rot
        glRotatef(x, 0, 1, 0)
        glRotatef(-y, math.cos(math.radians(x)), 0, math.sin(math.radians(x)))
        x, y, z = cam.pos
        glTranslatef(-x, -y, -z)
